/*
 * Program:   R  A  C
 * 
 * RAC - Rows and Columns, the text extraction tool 
 * http://aurelio.net/rac
 *
 * Copyright 2003  Aurelio Marinho Jargas
 *                 Thobias Salazar Trevisan
 *
 * Please address all bugs and comments to:
 *    verde@aurelio.net
 *    thobias@thobias.org
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, version 2.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You have received a copy of the GNU General Public License along
 *  with this program, on the COPYING file.
 *
 */

#include "rac.h"


/* function defines the LINE_START, LINE_END and STEP_LINE (maybe) 
 * address of a command
 */
void get_line_range(char *command)
{
	char *p1, *temp_line, c, *temp_p1;
	int i, j=0, n = strlen(command);

	p1 = xmalloc(strlen(command)+1, "get_line_range alloc p1.");
	temp_p1 = p1;
	p1[0] = '\0';
	/* erase between '[' and ']' 
	 * now, p1 has the addresss without [xxxxx]s */ 
	for (i=0; i < n; i++){
		if (command[i] != '[' && command[i] != ']' && j == 0)
			strncat(p1,&command[i],1);
		if (command[i] == '[') j = 1;
		if (command[i] == ']') j = 0;
	}
	/* we have some line address or just a '[xxxx]' address ?! */ 
	if (p1[0] == '\0') {
		rac.LINE_START =  1;
		rac.LINE_END   = -1;
		free(temp_p1);
		return;
	}
	/* to record the chunks (chars) of read line address */
	temp_line = xmalloc(strlen(p1)+1, "get_line_range alloc tem_line.");
	temp_line[0] = '\0';
#ifdef DEBUG2
	printf("--- only lines command = %s\n",p1);
#endif
	strncpy(&c,p1,1);
	while (c) {
		/* test what kind of address.
		 * it has line range */
		if (c == ':') {
			rac.LINE_START = temp_line[0] == '\0' ? 1 : atoi(temp_line);
			temp_line[0] = '\0';
			p1++;
			strncpy(temp_line,p1,strlen(p1));
			rac.LINE_END = p1[0] == '\0' ? -1 : atoi(p1);
			break;
		}
		/* line step */
		else if (c == '~') {
			rac.LINE_START = atoi(temp_line);
			if (rac.LINE_START == 0)
				error("", "line start can not be zero");

			temp_line[0] = '\0';
			p1++;
			strncpy(&c,p1,1);
			rac.STEP_LINE = !c ? 0 : atoi(p1);
			rac.LINE_END = rac.STEP_LINE < 0 ? 1 : -1;
			if (rac.STEP_LINE == 0)
				error("", "STEP_LINE can not be zero");
		}
		strncat(temp_line,p1,1);
		p1++;
		strncpy(&c,p1,1);
		/* one line only ex: 'x' */
		if (!c && rac.LINE_START == 0 && rac.LINE_END == 0)
				rac.LINE_START = rac.LINE_END = atoi(temp_line);
	}

#ifdef DEBUG2
	printf("--- line_start = %li line_end = %li STEP_LINE %li\n",rac.LINE_START,rac.LINE_END,rac.STEP_LINE);
#endif
	free(temp_line);
	free(temp_p1);
}

/* check and adjust start and end points 
 * its the same to lines and columns 
 * it returns -1 if its a invalida range 
 */
int check_range(long int *start, long int *end, long int total, int range_adjust)
{
	/* convert the negative address to positive */
	if (*start < 0) *start = total + (*start+1);
	if (*end < 0)   *end   = total + (*end+1);

	if (range_adjust == RANGE_ADJUST) {
		if (*start <= total && *start >= 1) 
			*end = *end < 1 ? 1 : (*end > total ? total : *end);
		if (*end <= total && *end >= 1) 
			*start = *start < 1 ? 1 : (*start > total ? total : *start);
	}

	return (*start > total || *start < 1 || *end > total || *end < 1 ? -1 : 1);
}

/* executes (print) commands type = TYPE_COL_RANGE
 * ex:  4[1,-1,4:9],  [4:8],  ...
 */
void exec_col_range(char *command, char *col_cmd_start, char *col_cmd_end)
{
	char *token, *temp_col, *p, *temp_line, *temp_step_col;
	long int i, j, t, print_newline = 0;
	long int col_start, col_end, temp_col_check;
	
	/* check and adapt the line number */
	if (check_range(&rac.LINE_START, &rac.LINE_END, rac.total_lines, RANGE_NOT_ADJUST) < 1)
		return;

#ifdef DEBUG2
	printf("--- Definitive Line Range = from %li to %li\n",rac.LINE_START,rac.LINE_END);
	printf("--- col start %s end %s\n",col_cmd_start, col_cmd_end);
#endif

	 /* if we have more than 1 line (ex: '[1,-1]), than we walk (to break the column
	  * command) 'n' (number of lines) times in the col_cmd_start memory */
	temp_col = xmalloc(strlen(col_cmd_start)+1,"exec_col_range alloc temp_col.");

	for (i=rac.LINE_START; i<=rac.LINE_END; i++) {
		/* locate and get the line */
		locate_line(i);
		temp_line = get_line(i);

		/* we can not count new line */
		t = temp_line[rac.num_chars_line[i-1]-1] == '\n' ? rac.num_chars_line[i-1]-1 : rac.num_chars_line[i-1];

		strncpy(temp_col,col_cmd_start,strlen(col_cmd_start));
		temp_col[strlen(col_cmd_start)] = '\0';
		/* break the command into separate commands */
		token = strtok(temp_col, ",");
		while (token != NULL) {
			rac.STEP_COL = 0;
			/* the column command has a specific STEP_COL */
			if ((temp_step_col=strchr(token,'~')) != NULL)
				rac.STEP_COL = atoi(++temp_step_col);
			/* it is a single char or STEP_COL range */
			if ((p = strchr(token,':')) == NULL) {
				col_start = atoi(token);
				temp_col_check = rac.STEP_COL > 0 ? -1 : 1;
				if (check_range(&col_start, &temp_col_check, t, RANGE_NOT_ADJUST) > 0) {
					if (rac.STEP_COL == 0)
						printf("%c",temp_line[col_start-1]);
					else
						for (j=col_start-1; j!=(temp_col_check-1)+rac.STEP_COL; j+=rac.STEP_COL) {
							/* as for uses != condition, we have to guarantee
							 * that we do not access away from the limits */
							if (col_start-1 < temp_col_check-1 && j > temp_col_check-1) break;
							if (col_start-1 > temp_col_check-1 && j < temp_col_check-1) break;
							printf("%c",temp_line[j]);
						}
					print_newline = 1;
				}
			}
			/* it is a column range */
			else {
				col_start = atoi(token) != 0 ? atoi(token) : 1;
				p++;
				if (token[0] == '0' && p[0] == '0') return;
				if ((!strcmp(token,"-0:-0")) || (!strcmp(token,"-0:0")) || (!strcmp(token,"0:-0")))
					return;
				if (p[0] == '0') col_end = 1;
				else col_end = atoi(p) != 0 ? atoi(p) : -1;
				if (check_range(&col_start, &col_end, t, RANGE_ADJUST) > 0) {
					if (rac.STEP_COL == 0) rac.STEP_COL = col_start <= col_end ? 1 : -1;
					for (j=col_start-1; j!=(col_end-1)+rac.STEP_COL; j+=rac.STEP_COL) {
						printf("%c",temp_line[j]);
					}
					print_newline = 1;
				}
			}
			/* get the next command */
			token = strtok(NULL, ",");
		}
		free(temp_line);
		if (print_newline == 1) {
			printf("\n");
			print_newline = 0;
		}
	}
	free(temp_col);
}

/* it executes (print) commands type = TYPE_LINE_RANGE
 * ex: 2,  1~2, 3:4,  3:4[2],  1[2]:4[-1],  1:-1,  ...  
 */
void exec_line_range(char *command, char *col_cmd_start, char *col_cmd_end)
{
	long int i, j, t;
	long int col_start, col_end, temp_col=1;
	char *temp_line;

	/* just one line or STEP_LINE is defined
	 * so we can not adjust the range */
	if (rac.LINE_START == rac.LINE_END || rac.STEP_LINE != 0) {
		if (check_range(&rac.LINE_START, &rac.LINE_END, rac.total_lines, RANGE_NOT_ADJUST) < 1) 
			return;
	}
	else
		if (check_range(&rac.LINE_START, &rac.LINE_END, rac.total_lines, RANGE_ADJUST) < 1) 
			return;

	/* define STEP_LINE, if it is not defined */
	if (rac.STEP_LINE == 0)
		rac.STEP_LINE = rac.LINE_START < rac.LINE_END ? 1 : -1;

#ifdef DEBUG2
	printf("--- Definitive Line Range = from %li to %li\n",rac.LINE_START,rac.LINE_END);
	printf("--- STEP_LINE = %li\n",rac.STEP_LINE);
	printf("--- col start %s end %s\n",col_cmd_start, col_cmd_end);
#endif

	for (i = rac.LINE_START; i != rac.LINE_END + rac.STEP_LINE; i += rac.STEP_LINE) {
		/* as for is i != LINE_END not < or > ... and we can have a STEP_LINE
		 * greater or less than 1. 
		 * so it could try to access a line away from the limits
		 * but we can permit this access 8=) */
		if (rac.LINE_START < rac.LINE_END && i > rac.LINE_END) 	break;
		if (rac.LINE_START > rac.LINE_END && i < rac.LINE_END) 	break;

		/* locate and get the line */
		locate_line(i);
		temp_line = get_line(i);

		/* we can not count new line */
		t = temp_line[rac.num_chars_line[i-1]-1] == '\n' ? rac.num_chars_line[i-1]-1 : rac.num_chars_line[i-1];

		if (i == rac.LINE_START && i == rac.LINE_END) {
			col_start = col_cmd_start == NULL ? 1 : atoi(col_cmd_start);
			col_end = col_cmd_end == NULL ? -1 : atoi(col_cmd_end);
			if (check_range(&col_start, &col_end, t, RANGE_ADJUST) < 1) {
				free(temp_line);
				break;
			}
			for (j=col_start-1;j<col_end;j++)
				printf("%c",temp_line[j]);
			printf("\n");
		}
		else if (i == rac.LINE_START) {
			col_start = col_cmd_start == NULL ? 1 : atoi(col_cmd_start);
			if (check_range(&col_start, &temp_col, t, RANGE_NOT_ADJUST) > 0) {
				for (j=col_start-1;j<rac.num_chars_line[i-1];j++)
					printf("%c",temp_line[j]);
				if (temp_line[j-1] != '\n') printf("\n");
			}
		}
		else if (i == rac.LINE_END) {
			col_end = col_cmd_end == NULL ? -1 : atoi(col_cmd_end);
			if (check_range(&temp_col, &col_end, t, RANGE_NOT_ADJUST) > 0) {
				for (j=0;j<col_end;j++)
					printf("%c",temp_line[j]);
				printf("\n");
			}
		}
		else printf("%s",temp_line);

		free(temp_line);
	}
}

/* separates the column command in start and end 
 * it do not analise the address, put in memory only the columns command
 * ex: col_cmd_start = 3:4,4,-1 col_cmd_end = 3
 */
void get_col_chunk(char *command, char **col_cmd_start, char **col_cmd_end)
{
	char c, *temp_col, *temp_command = command;
	int i=0, line_key = 0;

	temp_col = xmalloc(strlen(command)+1, "get_col_chunk alloc p1.");
	temp_col[0] = '\0';
	strncpy(&c,command,1);
	while (c) {
		if (c != '[' && i == 1) strncat(temp_col,command,1);
		else if (c == ':' && i == 0) line_key = 1; 	
		if (c == ']') {
			i = 0;
			/* erase the last char, ie, ']' */
			temp_col[strlen(temp_col)-1] = '\0';
			/* is it a start column command */
			if (line_key == 0) {
				*col_cmd_start = xmalloc(strlen(temp_col)+1,
						"get_col_chunk: alloc col_cmd_start.");
				strncpy(*col_cmd_start,temp_col,strlen(temp_col)+1);
			}
			/* it is a end column command */
			else {
				*col_cmd_end = xmalloc(strlen(temp_col)+1,
						"get_col_chunk: alloc col_cmd_end.");
				strncpy(*col_cmd_end,temp_col,strlen(temp_col)+1);
			}
			temp_col[0] = '\0';
		}
		if (c == '[') i = 1;
		command++;
		strncpy(&c,command,1);
	}
	/* do not have line range. ex: '[x,y]' address, so col_cmd_start 
	 * and col_cmd_end are equal */
	if (*col_cmd_start != NULL && line_key == 0) {
		*col_cmd_end = xmalloc(strlen(temp_col)+1,
				"get_col_chunk: alloc col_cmd_end.");
		strncpy(*col_cmd_end,*col_cmd_start,strlen(*col_cmd_start)+1);
	}
	command = temp_command;
#ifdef DEBUG2
	printf("--- col_cmd_start = %s\n",*col_cmd_start);
	printf("--- col_cmd_end   = %s\n",*col_cmd_end);
#endif
	free(temp_col);
}

/* execute the comands */
void exec_cmd(char *cmd_line)
{
	int i, k, type;
	char **command=NULL;
	/* it has the chunk of command that 
	 * has the columns address */
	char *col_cmd_start, *col_cmd_end;

	/* get the number of independent commands (separated by ';') */
	k = read_cmd(&command, convert_addressing(cmd_line));

	/* it executes independent commands */
	for (i=0;i<k;i++) {
#ifdef DEBUG2
		printf("--- Command = %s\n",command[i]);
#endif
		col_cmd_start = col_cmd_end = NULL;
		rac.LINE_START = rac.LINE_END = rac.STEP_LINE = 0;

		get_line_range(command[i]);
		get_col_chunk(command[i], &col_cmd_start, &col_cmd_end);
		type = get_command_type(command[i]);
		if (type == TYPE_LINE_RANGE)
			exec_line_range(command[i], col_cmd_start, col_cmd_end);
		else if (type == TYPE_COL_RANGE)
			exec_col_range(command[i], col_cmd_start, col_cmd_end);

		if (col_cmd_start != NULL) free(col_cmd_start);
		if (col_cmd_end   != NULL) free(col_cmd_end);
		free(command[i]);
	}
	free(command);
}


/* convert string and regular expression addresses
 * to line numbers addresses in the command line
 */
char *convert_addressing(char *cmd_line)
{
  struct str_conv_info c;
  char prev_char = 0, col_addr = 0, str_addr = 0;
  int i, line_size;

  line_size = strlen(cmd_line);
  cmd_line[line_size] = '\0';

  c.line_size = line_size + 1;
  c.conv_line = xmalloc(c.line_size, "convert_addressing alloc conv_line.");

  for (i = 0, c.curr_idx = 0; i < line_size; i++) {

    /* in column address */
    if (col_addr) {

      /* in string or RE address inside column address
       * for while this is not supported */
      if (str_addr) {

	/* skip backslash-escaped characters */
	if (cmd_line[i] == '\\' && cmd_line[i+1]) i++;
	else
	  /* exit string or RE address */
	  if (cmd_line[i] == str_addr) str_addr = 0;
      }
      else
	switch (cmd_line[i]) {
	  /* start string or RE address */
	case '=':
	case '/':
	  str_addr = cmd_line[i];
	  break;

	  /* exit column address */
	case ']':
	  prev_char = cmd_line[i];
	  col_addr = 0;
	  break;
	}
    }

    /* out of column, string and RE addresses */
    else {
      switch (cmd_line[i]) {
	/* skip blank characters */
      case ' ':
      case '\t':
	break;

	/* start column address */
      case '[':
	col_addr = 1;
	break;

	/* found the start of a string or RE address */
      case '=':
      case '/':
	if (prev_char && prev_char != ';' && prev_char != ':')
	  error("", "Invalid addressing.");

	/* convert string or RE address to line numbers addresses */
	i += convert_strings(&cmd_line[i], &c, prev_char);

	/* keep the last significant character */
      default:
	prev_char = cmd_line[i];
      }
    }

    /* realloc converted line if there is not enough space */
    if (c.curr_idx + 1 >= c.line_size) {
      c.line_size += NUM_STR_SIZE;
      c.conv_line = realloc(c.conv_line, c.line_size);
      if (!c.conv_line)
	error("convert_addressing", "realloc of converted line."); 
    }
    /* copy character that does not need conversion */
    c.conv_line[c.curr_idx++] = cmd_line[i];
  }
  c.conv_line[c.curr_idx] = '\0';

  return c.conv_line;
}


/* convert string or RE address to line numbers addresses
 * return the number of converted characters in command line
 */
int convert_strings(char *cmd_line, struct str_conv_info *c, char pc)
{
  struct str_addr_info str1, str2;
  int i;

  /* get first string and its address modifiers */
  i = get_addr_info(cmd_line, &str1);

  if (pc == ':' && !str1.occur)
    error("", "Ocorrncia da string no foi especificada.");

  /* there is a range modifier */
  if (cmd_line[i] == ':') {

    /* ops, double range is not allowed */
    if (pc == ':') error("", "Invalid Range.");
    str2.rac_str = 0;
    str2.occur = 1;

    /* skip blank characters */
    for (i += 1; cmd_line[i] == ' ' || cmd_line[i] == '\t'; i++);

    /* get second string and its address modifiers */
    if (cmd_line[i] == '=' || cmd_line[i] == '/')
      i += get_addr_info(&cmd_line[i], &str2);
    else
      /* do not convert if the first string does not exist in file */
      if (!str1.rac_str->num_occurs) {
	char str_addr = 0;

	for (i += 1; cmd_line[i]; i++)
	  if (str_addr) {
	    /* skip backslash-escaped characters */
	    if (cmd_line[i] == '\\' && cmd_line[i+1]) i++;
	    /* exit string or RE */
	    else if (cmd_line[i] == str_addr) str_addr = 0;
	  }
	  else {
	    /* reached a command separator */
	    if (cmd_line[i] == ';') break;
	    /* start string or RE */
	    else if (cmd_line[i] == '=' || cmd_line[i] == '/')
	      str_addr = cmd_line[i];
	  }
	return i;
      }

    /* both parts of the range must have or not a count occurrence modifier */
    if ((!str1.occur && str2.occur) || (str1.occur && !str2.occur))
      error("", "Ocorrncia da string no foi especificada.");

    /* convert the strings separately if each one addresses only one line */
    if (str1.occur && str2.occur) {

      /*** TODO --strict ***/
      convert_one_string(str1, c);
      c->conv_line[c->curr_idx++] = ':';
      if (str2.rac_str) convert_one_string(str2, c);
    }
    else {
      /* convert the range to multiples ranges doing
       * pars with str1 and str2 like sed */
      long int k1, k2, n1, n2;

      /*** TODO --mesma_linha ***/
      for (k1 = 0, k2 = -1; k1 < str1.rac_str->num_occurs &&
	     k2 < str2.rac_str->num_occurs; k1++) {
	n1 = str1.rac_str->occurs[k1] + str1.count;

	/* there is no need to continue if line number is after EOF */
	if (n1 > rac.total_lines) break;

	/* new par starts after previous par end */
	if (k2 < 0 || str1.rac_str->occurs[k1] > str2.rac_str->occurs[k2]) {
	  for (k2 += 1;
	       k2 < str2.rac_str->num_occurs &&
		 str2.rac_str->occurs[k2] <= str1.rac_str->occurs[k1];
	       k2++);

	  /*** TODO --strict ***/
	  if (k2 >= str2.rac_str->num_occurs) n2 = rac.total_lines + 1;
	  else n2 = str2.rac_str->occurs[k2] + str2.count;

	  /* discard par, because it is before BOF */
	  if (n1 < 1 && n2 < 1) continue;
	  if (n1 < 1) n1 = 1;
	  if (n2 < 1) n2 = 1;

	  /* add new line range in converted line */
	  add_num_conv(n1, str1, c);
	  c->conv_line[c->curr_idx++] = ':';
	  add_num_conv(n2, str2, c);
	  c->conv_line[c->curr_idx++] = ';';
	}
      }
    }
  }
  /*** TODO --strict (if pc is ':') ***/
  else convert_one_string(str1, c);

  return i;
}


/* convert only one string or RE address, ie,
 * a simple address or only one side of a range;
 * return false if string is not in file or the
 * resulting line number is before BOF or after EOF
 */
int convert_one_string(struct str_addr_info str, struct str_conv_info *c)
{
  int ret = 0;
  long int k, num_line;

  if (!str.rac_str) return 0;

  /* count occurrence modifier */
  if (str.occur) {

    /* negative occurrence */
    if (str.occur < 0) k = str.rac_str->num_occurs + str.occur;
    /* positive occurrence */
    else k = str.occur - 1;

    /* occurrence requested exists */
    if (k >=0 && k < str.rac_str->num_occurs) {
      /* sum count ahead modifier */
      num_line = str.rac_str->occurs[k] + str.count;

      /* line number is before BOF */
      if (num_line < 1) num_line = -(rac.total_lines + 1);
      /* line number is ok, before EOF */
      else if(num_line <= rac.total_lines) ret = 1;
    }
    /* requested does not exist, go out of file limits */
    else {
      if (str.occur > 0) num_line = rac.total_lines + 1;
      else num_line = -(rac.total_lines + 1);
    }

    /* add line address in converted line */
    add_num_conv(num_line, str, c);
  }
  /* all occurrences of string */
  else {
    for (k = 0; k < str.rac_str->num_occurs; k++) {

      /* sum count ahead modifier */
      num_line = str.rac_str->occurs[k] + str.count;

      if (num_line > 0 && num_line <= rac.total_lines) {
	ret = 1;

	/* add new line address in converted line */
	add_num_conv(num_line, str, c);
	c->conv_line[c->curr_idx++] = ';';
      }
    }
  }
  return ret;
}


/* add a line number in converted command line
 * that will grow if necessary
 */
void add_num_conv(long int num, struct str_addr_info str, struct str_conv_info *c)
{
  /* realloc converted line if there is not enough space */
  if (c->curr_idx + NUM_STR_SIZE > c->line_size) {
    c->line_size += NUM_STR_SIZE;
    c->conv_line = realloc(c->conv_line, c->line_size);
    if (!c->conv_line)
      error("add_num_conv", "realloc of converted line."); 
  }

  /* copy line number */
  c->curr_idx += sprintf(&c->conv_line[c->curr_idx], "%ld", num);

  /* there is step modifier */
  if (str.step_size) {

    /* realloc converted line if there is not enough space */
    if (c->curr_idx + str.step_size + 1 > c->line_size) {
      c->line_size = c->curr_idx + str.step_size + 1;
      c->conv_line = realloc(c->conv_line, c->line_size);
      if (!c->conv_line) 
	error("add_num_conv", "realloc of converted line."); 
    }

    /* copy step modifier */
    strncpy(&c->conv_line[c->curr_idx], str.step, str.step_size);
    c->curr_idx += str.step_size;
  }

  /* there is column complement */
  if (str.col_size) {

    /* realloc converted line if there is not enough space */
    if (c->curr_idx + str.col_size + 1 > c->line_size) {
      c->line_size = c->curr_idx + str.col_size + 1;
      c->conv_line = realloc(c->conv_line, c->line_size);
      if (!c->conv_line)
	error("add_num_conv", "realloc of converted line."); 
    }

    /* copy column complement */
    strncpy(&c->conv_line[c->curr_idx], str.col, str.col_size);
    c->curr_idx += str.col_size;
  }
}


/* fill the struct with the information about string address
 * return the size of string address
 */
int get_addr_info(char *addr, struct str_addr_info *info)
{
  int i, n, str_end, col_start;
  char delim = addr[0], isRE = (delim == '/' ? 1 : 0);
  char str_in_col = 0, num_str[NUM_STR_SIZE];
  regex_t re;
  regmatch_t match[7];	/* size depends on number of () in ER_STR_MODIFIERS */

  for (i = 1; addr[i]; i++) {

    /* skip backslash-escaped characters */
    if (addr[i] == '\\' && addr[i+1]) i++;
    else
      /* found the end of string or RE */
      if (addr[i] == delim) {

	/* search for string or RE in rac.strings */
	addr[i] = '\0';
	info->rac_str = rac.strings;
	while (info->rac_str) {
	  if (info->rac_str->isRE == isRE
	      && !strcmp(&addr[1], info->rac_str->string)) break;
	  info->rac_str = info->rac_str->next;
	}
	if (!info->rac_str) 
	  error("get_addr_info", "unexpected error searching string.");
	addr[i] = delim;
	str_end = i + 1;

	/* get address modifiers */
	if (regcomp(&re, ER_STR_MODIFIERS, REG_EXTENDED) != 0)
	  error("get_addr_info", "regular expression for string modifiers.");
	regexec(&re, &addr[str_end], 7, match, 0);

	/* match groups:
	 * 0 = entire modifiers matched
	 * 1 = count occurrence
	 * 2 = count negative occurrence
	 * 3 = number (count occurrence)
	 * 4 = count ahead or back
	 * 5 = number (count ahead)
	 * 6 = step or negative step */

	/* for while step without count occurrence is forbidden */
	if (match[6].rm_so != -1 && match[1].rm_so == -1)
	  error("", "A ocorrncia inicial deve ser especificada.");

	/* check for column complement after modifiers */
	i = str_end + match[0].rm_eo;
	col_start = i;
	if (addr[i] == '[') {

	  info->col = &addr[i];
	  for (i += 1; addr[i]; i++) {

	    /* in string or RE address inside column complement
	     * for while this is not supported */
	    if (str_in_col) {

	      /* skip backslash-escaped characters */
	      if (addr[i] == '\\' && addr[i+1]) i++;
	      else
		/* exit string or RE address */
		if (addr[i] == str_in_col) str_in_col = 0;
	    }
	    else {
	      if (addr[i] == '=' || addr[i] == '/')
		/* start string or RE address */
		str_in_col = addr[i];
	      else
		if (addr[i] == ']')
		  /* found the end of column complement */
		  break;
	    }
	  }
	  /* skip tail blank characters */
	  for (i += 1; addr[i] == ' ' || addr[i] == '\t'; i++);
	}
	else info->col = NULL;
	info->col_size = i - col_start;

	if (addr[i] && addr[i] != ':' && addr[i] != ';')
	  error("", "Invalid addressing.");

	/* there is count ahead (or back) modifier */
	if (match[4].rm_so != -1) {

	  n = match[5].rm_eo - match[5].rm_so;
	  if (n < NUM_STR_SIZE) {
	    strncpy(num_str, &addr[str_end + match[5].rm_so], n);
	    num_str[n] = '\0';
	  }
	  else {
	    strncpy(num_str, &addr[str_end + match[5].rm_so], NUM_STR_SIZE-1);
	    num_str[NUM_STR_SIZE-1] = '\0';
	  }
	  /* count ahead */
	  info->count = atol(num_str);

	  /* count back */
	  if (addr[str_end + match[4].rm_so] == '-') 
	    info->count = -info->count;
	}
	else info->count = 0;

	/* there is count occurrence (or negative) modifier */
	if (match[1].rm_so != -1) {

	  n = match[3].rm_eo - match[3].rm_so;
	  if (n < NUM_STR_SIZE) {
	    strncpy(num_str, &addr[str_end + match[3].rm_so], n);
	    num_str[n] = '\0';
	  }
	  else {
	    strncpy(num_str, &addr[str_end + match[3].rm_so], NUM_STR_SIZE-1);
	    num_str[NUM_STR_SIZE-1] = '\0';
	  }
	  /* count occurrence */
	  info->occur = atol(num_str);

	  /* count negative occurence */
	  if (match[2].rm_eo - match[2].rm_so)
	    info->occur = -info->occur;

	  /* zero is a special case, it is not a valid occurrence
	   * so go far way to be different from no occurrence */
	  if (info->occur == 0) info->occur = -(rac.total_lines + 1);
	}
	else info->occur = 0;

	/* there is step modifier */
	if (match[6].rm_so != -1) {
	  info->step = &addr[str_end + match[6].rm_so];
	  info->step_size = match[6].rm_eo - match[6].rm_so;
	}
	else {
	  info->step = NULL;
	  info->step_size = 0;
	}

	/* mission accomplished */
	break;
      }
  }
  return i;
}
