/*
 * Program:   R  A  C
 * 
 * RAC - Rows and Columns, the text extraction tool 
 * http://aurelio.net/rac
 *
 * Copyright 2003  Aurelio Marinho Jargas
 *                 Thobias Salazar Trevisan
 *
 * Please address all bugs and comments to:
 *    verde@aurelio.net
 *    thobias@thobias.org
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, version 2.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You have received a copy of the GNU General Public License along
 *  with this program, on the COPYING file.
 *
 */

#include "rac.h"


/* get the number of lines and
 * number of chars per line 
 */
void get_input_info()
{
	long int bytes_read, vector_pos, buf_len;
	char buf[BUFFER_SIZE];

	bytes_read = vector_pos = buf_len = 0;
	while (fgets(buf, sizeof(buf), rac.stream) != NULL) {
		buf_len = strlen(buf);
		/* test if we have a new line */
		if (buf[buf_len - 1] == '\n') { 
			rac.total_lines++;
			vector_pos++;
			rac.num_chars_line = realloc(rac.num_chars_line, vector_pos * sizeof(long int));
			if (!rac.num_chars_line)
				error("get_input_info", "vector alloc of chars per line");
			if (((bytes_read + buf_len)-LONG_MAX) > 0)
				error("get_input_info", "line %li overflow LONG_MAX",rac.total_lines);
			
			bytes_read += buf_len;
			rac.num_chars_line[rac.total_lines - 1] = bytes_read;
#ifdef DEBUG2
			printf("--- line = %li size = %li\n",rac.total_lines,bytes_read);
#endif
			bytes_read = 0;
			search_strings(buf);
		}
		else {
			if (((bytes_read + buf_len)-LONG_MAX) > 0)
				error("get_input_info", "line %li overflow LONG_MAX",rac.total_lines);
			bytes_read += buf_len;
		}
		/* if input comes from stdin
		 * record the buf in a temp file */
		if (rac.stream == stdin) 
			if ((fwrite(buf, sizeof (char), buf_len, rac.tmp)) < strlen(buf)) 
				error("get_input_info", "can not write in the temp file");
	}
	/* text file might not end with a newline 
	 * like 'echo -n test' */
	if (buf_len && buf[buf_len - 1] != '\n') { 
		rac.total_lines++;
		vector_pos++;
		rac.num_chars_line = realloc(rac.num_chars_line, vector_pos * sizeof(long int));
		if (!rac.num_chars_line)
			error("get_input_info", "Error: vector alloc of chars per line");
		rac.num_chars_line[rac.total_lines - 1] = bytes_read;
		/* if input comes from stdin
		 * record the buf in a temp file */
		if (rac.stream == stdin) 
			if ((fwrite(buf, sizeof (char), buf_len, rac.tmp)) < strlen(buf))
				error("get_input_info", "can not write in the temp file");
#ifdef DEBUG2
		printf("--- line = %li size = %li\n",rac.total_lines,bytes_read);
#endif
		search_strings(buf);
	}
	/* test if we read the whole file */
	if (!feof(rac.stream)) 
		error("get_input_info", "Error reading from input");
	/* if we have stdin, switch between stream and tmp,
	 * so we can keep using the variable stream */
	if (rac.stream == stdin) {
		if (fclose(rac.stream))
			error("get_input_info", "Error: File could not be closed");
		rac.stream = rac.tmp;
	}
}


/* create a random temp file in the DEFAULT_TMPDIR or
 * remove the temp file generated by rac
 * it is used if the input comes from stdin 
 */
void temp_file(int opt)
{
	static char *tempfile = NULL;
	int fd;

	if (opt == REMOVE_TEMP_FILE) {
		if (remove(tempfile)) {
			fprintf(stderr, "Error: could not remove %s file.\n",tempfile);
			exit(1);
		}
#ifdef DEBUG2
		else
			printf("--- removed temp file %s\n",tempfile);
#endif
		free(tempfile);
		return;
	}
	if (opt == CREATE_TEMP_FILE) {
		tempfile = xmalloc(strlen (DEFAULT_TMPDIR) + strlen(TMPFILE_SUFIX) + 8,
						"temp_file function");
		/* create a random temp file */
		strcpy(tempfile,DEFAULT_TMPDIR);
		strncat(tempfile,TMPFILE_SUFIX,strlen(TMPFILE_SUFIX));
		sprintf (tempfile, "%sXXXXXX", tempfile);
		if ((fd = mkstemp (tempfile)) == -1) {
			fprintf(stderr, "Error: mkstemp to file %s.\n",tempfile);
			exit(1);
		}

		if (!(rac.tmp = fdopen(fd, "w+"))) {
			fprintf(stderr, "Error: could not create %s file.\n",tempfile);
			exit(1);
		}
#ifdef DEBUG1
		printf("--- created temp file %s\n",tempfile);
#endif
	}
}


/* search for rac.strings into the input current line
 * keep note of the line number for each string found
 */
void search_strings(char *line)
{
	struct address_string *str;
	char *buf, *nl = NULL;
	int found;

	/* new line must be removed to work regex with $ */
	nl = strchr(line, '\n');
	if (nl) *nl = '\0';

	if (icase) {
	  int i, n = strlen(line);
	  buf = xmalloc(n+1, "search_strings alloc buf.");
	  for (i=0; i < n; i++) 
	    buf[i] = tolower(line[i]);
	}
	else buf = line;
	str = rac.strings;
	while (str) {
		if (str->isRE) {
		  int re_err;
		  regex_t reg;

		  re_err = regcomp(&reg , str->escaped, ER_FLAGS);
		  if (re_err) {
		    char *err_buf;
		    size_t err_len = regerror(re_err, &reg, 0, 0);

		    err_buf = xmalloc(err_len, "search_strings alloc err_buf.");
		    regerror(re_err, &reg, err_buf, err_len);
		    error("", err_buf);
		  }
		  found = !regexec(&reg, buf, 0, 0, 0);
		}
		else found = (int) strstr(buf, str->escaped);

		/*** TODO --mesma_linha ***/
		if (found) {
			if (str->num_occurs % OCCURS_SIZE == 0) {
				str->occurs = realloc(str->occurs, (str->num_occurs + OCCURS_SIZE) * sizeof(long int));
				if (!str->occurs) 
					error("search_strings", "vector alloc of string occurrences.");
			}
			str->occurs[str->num_occurs++] = rac.total_lines;
		}
		str = str->next;
	}
	if (icase) free(buf);
	if (nl) *nl = '\n';
}
