/*
 * Program:   R  A  C
 * 
 * RAC - Rows and Columns, the text extraction tool 
 * http://aurelio.net/rac
 *
 * Copyright 2003  Aurelio Marinho Jargas
 *                 Thobias Salazar Trevisan
 *
 * Please address all bugs and comments to:
 *    verde@aurelio.net
 *    thobias@thobias.org
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, version 2.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You have received a copy of the GNU General Public License along
 *  with this program, on the COPYING file.
 *
 */

#include "rac.h"


void usage(int status)
{
	if (status !=0)
		fprintf (stderr, "Tente `%s --help' para mais informaes.\n",PROGRAM_NAME);
	else {
		fprintf(stdout,"\
Uso: %s ENDEREO [ARQUIVO]\n\
\n\
    Linha     Coluna      Significado do endereo\n\
    ----------------------------------------------------------------\n\
      N         [N]       N\n\
    N;M;I     [N,M,I]     N, M e I\n\
     N:M       [N:M]      Trecho de N at M\n\
     N~M       [N~M]       partir de N, ande de M em M\n\
    ----------------------------------------------------------------\n\
",PROGRAM_NAME);
	fprintf(stdout,"\
     =T=         -        Linhas que contm o texto T\n\
     =T=+N       -        N-sima linha aps a que contm T\n\
     =T=-N       -        N-sima linha anterior a que contm T\n\
     =T=*N       -        N-sima linha que contm T\n\
     =T=*N~M     -        N-sima linha com T, ento ande de M em M\n\
\n\
* Se o nmero do endereo for negativo, conta  partir do fim.\n\
* Se omitido o nome do arquivo ou usado '-', o texto ser lido\n\
  da entrada padro (STDIN).\n\
* Mais informaes na pgina do programa: %s\n\n",HTTP_ADDRESS);
	}
	exit (status == 0 ? 0 : 1);
}


int main(int argc, char *argv[])
{
	int i, n, addr_pos = 1, file_pos = 0;

	if (argc < 2)
		usage(1);

	/* default flags used in regcomp. (user ER) */
	ER_FLAGS = REG_EXTENDED | REG_NOSUB;
	icase = 0;
	/* ugly code, i know
	 * we can not use getopt, getopt is not portable */
	while (argc > addr_pos) {
		if ((!strcmp(argv[addr_pos],"-h")) || (!strcmp(argv[addr_pos],"--help")))
			usage(0);
		else if (!strcmp(argv[addr_pos],"-V") || !strcmp(argv[addr_pos],"--version")) {
			fprintf(stdout, "%s verso %s\n", PROGRAM_NAME,VERSION);
			exit(0);
		}
		/* default is Extended Regular Expression syntax 
		 * -b or --bre switch to Basic Regular Expression syntax */ 
		else if ((!strcmp(argv[addr_pos],"-b")) || (!strcmp(argv[addr_pos],"--bre")))
			ER_FLAGS = REG_NOSUB;
		else if ((!strcmp(argv[addr_pos],"-bi")) || (!strcmp(argv[addr_pos],"-ib"))) {
			ER_FLAGS = REG_NOSUB;
			icase = 1;
		}
		/* ignore case */
		else if ((!strcmp(argv[addr_pos],"-i")) || (!strcmp(argv[addr_pos],"--ignore-case")))
			icase = 1;
		/* basic testing for invalid option. TODO: improve me */
		else if ((argv[addr_pos][0] == '-')  && ((argv[addr_pos][1] < 48) || (argv[addr_pos][1] > 57)))
			error("", "Opo %s invlida",argv[addr_pos]);
		else 
			break;
		addr_pos++;
	}

	/* is there a file to open ?! */
	if (argc > addr_pos+1) { 
		if ((!strcmp(argv[addr_pos+1],"--") && (argc > addr_pos+2)) && (strcmp(argv[addr_pos+2],"-")))
			file_pos = addr_pos + 2;
		else if ((strcmp(argv[addr_pos+1],"-")) && (strcmp(argv[addr_pos+1],"--")))
			file_pos = addr_pos + 1;
	}

	bzero(&rac,sizeof(rac));

	/*convert to lower case */
	if (icase == 1) {
		n = strlen(argv[addr_pos]);
		for (i=0; i < n; i++)
			argv[addr_pos][i] = tolower(argv[addr_pos][i]);
	}
	
	/* test if we have to read from a file or stdin */
	if (!file_pos) {
#ifdef DEBUG1
		printf("--- Reading from stdin\n");
#endif
		rac.stream = stdin;
		temp_file(CREATE_TEMP_FILE);
		/* garbage collection using a unix feature :) 
		 *
		 * if its reading from stdin, the rac will create a temp file
		 * that temp file was created and opened in the above function
		 * so we can remove this file, because the process already have
		 * the pointer to the file descriptor. the process will only
		 * close the file at the end
		 *
		 * we really have to test this technique */
		temp_file(REMOVE_TEMP_FILE);
	}
	else {
		if (!(rac.stream = fopen(argv[file_pos], "r")))
			error("", "Erro ao abrir o arquivo %s", argv[file_pos]);
	}
	
	rac.total_lines = 0;
	rac.num_chars_line = NULL;

	extract_strings(argv[addr_pos]);
	
	/* get the input informations, ie, total number of line and
	 * the number of chars per line. used to 'fseek' a specific line */
	get_input_info();
#ifdef DEBUG1
	printf("--- Total lines = %li\n",rac.total_lines);
#endif

	/* execute the commands */
	exec_cmd(argv[addr_pos]);

	free(rac.num_chars_line);
	if (fclose(rac.stream))
		error("main", "Erro ao fechar o arquivo");

	exit(0);
}


