/*
 * Program:   R  A  C
 * 
 * RAC - Rows and Columns, the text extraction tool 
 * http://aurelio.net/rac
 *
 * Copyright 2003  Aurelio Marinho Jargas
 *                 Thobias Salazar Trevisan
 *
 * Please address all bugs and comments to:
 *    verde@aurelio.net
 *    thobias@thobias.org
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, version 2.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You have received a copy of the GNU General Public License along
 *  with this program, on the COPYING file.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <limits.h>
#include <string.h>
#include <time.h>
#include <ctype.h>
#include <stdarg.h>
#include <regex.h>

#define PROGRAM_NAME "rac"

#define VERSION "1.3"

#define AUTHORS "Aurelio Marinho Jargas, Thobias Salazar Trevisan"

#define BUG_ADDRESS "verde@aurelio.net, thobias@thobias.org"

#define HTTP_ADDRESS "http://aurelio.net/rac"

#ifndef LONG_MAX
#define LONG_MAX	2147483647L
#endif

#ifndef INT_MAX
#define INT_MAX		2147483647
#endif

/* regular expression to identify the command type */
#define ER_TYPE_RANGE_COL    "^-?[0-9]*\\[([0-9:,~-]+)+\\]$"
#define ER_TYPE_RANGE_LINE   "^((-?[0-9]*(\\[[0-9-]*\\])?:-?[0-9]*(\\[[0-9-]*\\])?)|(-?[0-9]+~?-?[0-9]*))$"
#define ER_STR_MODIFIERS "^[[:space:]]*(\\*[[:space:]]*(-?)[[:space:]]*([0-9]+))?[[:space:]]*([+-][[:space:]]*([0-9]+))?[[:space:]]*(~[[:space:]]*-?[[:space:]]*[0-9]+)?[[:space:]]*"

/* type of command */
#define TYPE_LINE_RANGE 0
#define TYPE_COL_RANGE  1

/* used to call the temp_file function */
#define CREATE_TEMP_FILE 0
#define REMOVE_TEMP_FILE 1

#ifndef DEFAULT_TMPDIR
# define DEFAULT_TMPDIR "/tmp"
#endif

/* sufix to create the temp file (if necessary) */
#define TMPFILE_SUFIX "/rac_temp_"

/* what kind of adjustment will be done at check_range function */
#define RANGE_ADJUST      0
#define RANGE_NOT_ADJUST  1

/* command allocate memory (step) */
#define CMD_SIZE	32

/* the number of bytes per atomic read */
/* BUFFER_SIZE can not be 1. fgets reads size - 1*/
#define BUFFER_SIZE 8192

/* flags to "regcomp" the user ER */
int ER_FLAGS;

/* ignore case or not */
int icase;

/* maximum size of error message */
#define MAX_ERROR_SIZE  256

/* number of chars to keep a long int followed by a ; and a '\0' */
#define NUM_STR_SIZE 12

/* size block to allocate occurs vetor */
#define OCCURS_SIZE 10

struct address_string {
	char *string;
	char *escaped;
	char isRE;
	long int *occurs;
	long int num_occurs;
	struct address_string *next;
};

/* store string address information
 * used to convert string addresses to line numbers
 */
struct str_addr_info {
  struct address_string *rac_str;
  /* address modifiers */
  long int count;	/* count ahead or back */
  long int occur;	/* count occurrence */
  char *step;		/* step foward or back */
  int step_size;	/* size of string with step */
  char *col;		/* column complement */
  int col_size;		/* size of string with column */
};

/* converted command line information */
struct str_conv_info {
  char *conv_line;
  int curr_idx;
  int line_size;
};

struct struct_rac {
	/* vector used to seek lines in the input and
	 * to know how many chars each line has */
	long int *num_chars_line; 

	/* the input's total lines number*/
	long int total_lines;

	/* range line start and end */
	long int LINE_START, LINE_END;

	/* range column start and end */
	long int COL_START, COL_END;

	/* step line and step col */
	long int STEP_LINE, STEP_COL;

	FILE *stream;
	FILE *tmp;
	struct address_string *strings;
};

struct struct_rac rac;

/* Function Prototyping */
void *xmalloc(size_t n, char string[]);		 /* lib_rac.c */
void error(char *func, char *fmt, ...);			 /* lib_rac.c */
void locate_line(long int num_line);         /* lib_rac.c */
char *get_line(long int line);				 /* lib_rac.c */
int get_command_type(char *command);		 /* lib_rac.c */
int read_cmd(char ***command, char *line); /* lib_rac.c */
void get_input_info(void);    /* input.c */
void temp_file(int opt);      /* input.c */
void get_line_range(char *command);  /* exec.c */
void exec_line_range(char *command, char *col_cmd_start, char *col_cmd_end); /* exec.c */
int get_command_type(char *command); /* exec.c */
int check_range(long int *start, long int *end, long int total, int range_adjust); /* exec.c */
void get_col_chunk(char *command, char **col_cmd_start, char **col_cmd_end); /* exec.c */
void exec_cmd(char *cmd_line);         /* exec.c */
void usage(int status);            /* rac.c */
int main(int argc, char *argv[]);  /* rac.c */
void extract_strings(char *addrs);           /* lib_rac.c */
void escape_string(char *str, char delim);   /* lib_rac.c */
void add_string(char *str, char isRE);       /* lib_rac.c */
void search_strings(char *line);             /* input.c */
char *convert_addressing(char *line);        /* exec.c */
int convert_strings(char *cmd_line, struct str_conv_info *c, char pc);  /* exec.c */
int convert_one_string(struct str_addr_info str, struct str_conv_info *c);  /* exec.c */
void add_num_conv(long int num, struct str_addr_info str, struct str_conv_info *c);  /* exec.c */
int get_addr_info(char *addr, struct str_addr_info *info);  /* exec.c */

